@@ -636,6 +636,7 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 {
 	gnutls_x509_crt_t xcrt;
 	gnutls_x509_privkey_t xkey;
+	gnutls_pubkey_t xpubkey;
 	int ret;
 	gnutls_datum_t *secret_key;
 	unsigned key_usage = 0;
@@ -706,9 +707,22 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 		}
 	}
 
-	if (xkey == NULL && xcrt == NULL && secret_key == NULL) {
+	xpubkey = load_pubkey(0, info);
+	if (xpubkey != NULL) {
+		ret =
+		    gnutls_pkcs11_copy_pubkey(url, xpubkey, label,
+						     &cid,
+						     0, flags);
+		if (ret < 0) {
+			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
+				__LINE__, gnutls_strerror(ret));
+			exit(1);
+		}
+	}
+
+	if (xkey == NULL && xcrt == NULL && secret_key == NULL && xpubkey == NULL) {
 		fprintf(stderr,
-			"You must use --load-privkey, --load-certificate or --secret-key to load the file to be copied\n");
+			"You must use --load-privkey, --load-certificate, --load-pubkey or --secret-key to load the file to be copied\n");
 		exit(1);
 	}
 
